import random
from dataclasses import dataclass
from functools import cached_property
from typing import Iterable, List, Type

from tilted.constants import (
    NUM_BITS,
    ROYAL_FLUSH_VALUE,
    STRAIGHT_VALUE,
    WHEEL_STRAIGHT_VALUE,
)
from tilted.enums import CardRank, CardSuit, HandRank
from tilted.exceptions import EmptyDeckException
from tilted.hand_evaluation import HandEvaluator
from tilted.utils import bit_sequence_to_int, get_binary_index_from_card_rank


@dataclass(frozen=True)
class Card:
    rank: CardRank
    suit: CardSuit

    @staticmethod
    def sort(card):
        return card.rank.value

    @cached_property
    def value(self):
        return self.rank.value + 1

    def __eq__(self, other):
        return self.rank.value == other.rank.value

    def __repr__(self):
        return f"{CardRank.as_str(self.rank)}{self.suit.value}"


@dataclass
class Hand:
    cards: Iterable[Type[Card]]

    def __post_init__(self):
        assert len(self.cards) == 5, "A poker hand is 5 cards."
        assert len(self.cards) == len(
            set(self.cards)
        ), "All cards in a hand must be unique."
        self.cards = sorted(self.cards, key=Card.sort)
        self.evaluator = HandEvaluator(self)

    def __repr__(self) -> str:
        cards = " ".join(str(card) for card in self.cards)
        return f"<Hand: {cards}>"

    def __gt__(self, other) -> bool:
        is_same_hand_rank = self.evaluator.is_same_hand_rank(other)
        is_better_hand = self.hand_rank.value > other.hand_rank.value
        return (
            self.evaluator.break_tie(other) == self
            if is_same_hand_rank
            else is_better_hand
        )

    def __lt__(self, other) -> bool:
        is_same_hand_rank = self.evaluator.is_same_hand_rank(other)
        is_worse_hand = self.hand_rank.value < other.hand_rank.value
        return (
            self.evaluator.break_tie(other) is other
            if is_same_hand_rank
            else is_worse_hand
        )

    def __eq__(self, other) -> bool:
        is_same_hand_rank = self.evaluator.is_same_hand_rank(other)
        return self.evaluator.break_tie(other) is None if is_same_hand_rank else False

    def __contains__(self, value: CardRank) -> bool:
        return any([card.rank == value for card in self.cards])

    @cached_property
    def card_values(self) -> int:
        """
        Returns int representation of a binary number of the form:

        A K Q J T 9 8 7 6 5 4 3 2 - -
        1 1 1 1 1 0 0 0 0 0 0 0 0 0 0

        Where a 1 represents that the card rank is in the hand and a 0
        represents that the card rank isn't in the hand.
        """
        result = list("0" * NUM_BITS)
        for card in self.cards:
            binary_idx = get_binary_index_from_card_rank(card.rank)
            result[binary_idx] = "1"
        return bit_sequence_to_int(result)

    @cached_property
    def card_counts(self) -> int:
        """
        Returns int representation of a binary number of the form:

        A    K    Q    J    T    9    8    7    6    5    4    3    2    -    -
        0001 0001 0001 0001 0001 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000

        Where each 4-bit field represents the number of cards of that rank
        in the hand.
        """
        result = ["0000" for _ in range(NUM_BITS)]
        counts = {
            card.rank: len(
                [
                    count_card
                    for count_card in self.cards
                    if count_card.rank == card.rank
                ]
            )
            for card in self.cards
        }
        for rank, count in counts.items():
            binary_idx = get_binary_index_from_card_rank(rank)
            binary_value = ("1" * count).zfill(4)
            result[binary_idx] = binary_value
        return bit_sequence_to_int(result)

    @cached_property
    def is_straight(self) -> bool:
        lsb = self.card_values & -self.card_values
        return (
            self.card_values / lsb == STRAIGHT_VALUE
            or self.card_values == WHEEL_STRAIGHT_VALUE
        )

    @cached_property
    def is_wheel(self) -> bool:
        """
        In poker, the term "wheel" refers to the ace-low straight (A2345).
        """
        return self.is_straight and CardRank.ACE in self and CardRank.FIVE in self

    @cached_property
    def is_flush(self) -> bool:
        return len(set(card.suit for card in self.cards)) == 1

    @cached_property
    def is_royal_flush(self) -> bool:
        return (
            self.is_straight and self.is_flush and self.card_values == ROYAL_FLUSH_VALUE
        )

    @cached_property
    def hand_rank(self) -> HandRank:
        return self.evaluator.get_hand_rank()


class Deck:
    cards: List[Card]

    def __init__(self):
        self.cards = []

        for suit in CardSuit:
            for rank in CardRank:
                self.cards.append(Card(rank, suit))

        random.shuffle(self.cards)

    def __len__(self) -> int:
        return len(self.cards)

    def draw(self) -> Card:
        try:
            return self.cards.pop()
        except IndexError as ex:
            raise EmptyDeckException from ex

    def draw_many(self, num_cards: int) -> List[Card]:
        return [self.draw() for _ in range(num_cards)]

    def burn(self) -> None:
        self.draw()



from enum import Enum, unique


@unique
class CardSuit(Enum):
    CLUBS = "♣"
    DIAMIONDS = "♦"
    HEARTS = "♥"
    SPADES = "♠"


@unique
class CardRank(Enum):
    TWO = 1
    THREE = 2
    FOUR = 3
    FIVE = 4
    SIX = 5
    SEVEN = 6
    EIGHT = 7
    NINE = 8
    TEN = 9
    JACK = 10
    QUEEN = 11
    KING = 12
    ACE = 13

    @staticmethod
    def as_str(entry):
        return str(
            entry.name[0] if entry.value > CardRank.NINE.value else entry.value + 1
        )


@unique
class HandRank(Enum):
    HIGH_CARD = 1
    PAIR = 2
    TWO_PAIR = 3
    THREE_OF_A_KIND = 4
    STRAIGHT = 5
    FLUSH = 6
    FULL_HOUSE = 7
    FOUR_OF_A_KIND = 8
    STRAIGHT_FLUSH = 9
    ROYAL_FLUSH = 10



from typing import Any, Sequence

from tilted.constants import NUM_BITS, UNUSED_BITS
from tilted.enums import CardRank


def combinations(items: Sequence[Any], n: int):
    if n == 1:
        for x in items:
            yield [x]
    for i in range(len(items)):
        for x in combinations(items[:i], n - 1):
            yield [items[i]] + x


def bit_sequence_to_int(bit_list: Sequence[str]):
    bits = "".join(bit_list)
    return int(f"0b{bits}", 2)


def get_binary_index_from_card_rank(rank: CardRank):
    return NUM_BITS - rank.value - UNUSED_BITS



from collections import Counter
from dataclasses import dataclass
from functools import reduce
from operator import ior
from typing import TYPE_CHECKING, List, Type

from tilted.constants import (
    BINARY_ARITHMETIC_RESULT_TO_HAND_RANK,
    NUM_BITS,
    TIEBREAKER_BITWISE_LEFT_SHIFTS,
)
from tilted.enums import HandRank

if TYPE_CHECKING:
    from tilted.models import Card, Hand


@dataclass
class HandEvaluator:
    hand: Type["Hand"]

    def _get_cards_ordered_by_frequency_and_face_value(self) -> List[Type["Card"]]:
        cards = self.hand.cards
        counter = Counter([card.value for card in cards])
        return sorted(cards, key=lambda card: (-counter[card.value], card.value))

    def get_hand_rank(self) -> HandRank:
        modulo_division_result = self.hand.card_counts % NUM_BITS  # type: ignore
        rank = BINARY_ARITHMETIC_RESULT_TO_HAND_RANK[modulo_division_result]

        if rank == HandRank.HIGH_CARD:
            # If binary result is high card, check for
            # hands which can't be determined by modulo division
            if self.hand.is_flush:
                if self.hand.is_straight:
                    if self.hand.is_royal_flush:
                        return HandRank.ROYAL_FLUSH
                    return HandRank.STRAIGHT_FLUSH
                return HandRank.FLUSH
            elif self.hand.is_straight:
                return HandRank.STRAIGHT
        return rank

    def get_tiebreak_score(self):
        """
        In order to break ties, we sort cards by their frequency of occurence
        and rank, then left shift each card's rank by TIEBREAKER_BITWISE_LEFT_SHIFTS.
        """
        ordered_cards = self._get_cards_ordered_by_frequency_and_face_value()

        left_shifted_bits = [
            card.value << shift_by
            for card, shift_by in zip(ordered_cards, TIEBREAKER_BITWISE_LEFT_SHIFTS)
        ]

        score = reduce(ior, left_shifted_bits)
        return score

    def is_same_hand_rank(self, other) -> bool:
        return self.get_hand_rank() == other.evaluator.get_hand_rank()

    def break_tie(self, other_hand: Type["Hand"]) -> Type["Hand"] | None:
        score = self.get_tiebreak_score()
        other_score = other_hand.evaluator.get_tiebreak_score()
        if score == other_score:
            return None

        # If either hand is a wheel (low straight A2345) then the higher
        # straight wins. Wheel vs wheel would have the same score and be
        # handled above.
        either_hand_is_wheel = any([self.hand.is_wheel, other_hand.is_wheel])
        if either_hand_is_wheel:
            winner = next(hand for hand in (self.hand, other_hand) if not hand.is_wheel)
            return winner

        # Return the hand with the highest score
        return self.hand if score > other_score else other_hand



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tilted']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'tilted',
    'version': '0.1.2',
    'description': 'Tilted is a lightweight, open-source Python package with a simple interface for poker hand evaluation & comparison.',
    'long_description': '\n# Tilted\n<img align="left" src="https://user-images.githubusercontent.com/8881202/169894189-c4d64c08-7751-4d0e-a95f-640f07c2e7bd.jpeg" width="100" height="100" />\n\nTilted is a lightweight, open-source Python package with a simple interface for poker hand evaluation & comparison.\n\n<img src="https://img.shields.io/github/v/release/MaxAtkinson/tilted" />\n\n<br />\n<br />\n\n## Installation\nWith Poetry:\n```sh\npoetry add tilted\n```\n\nWith Pip:\n```sh\npip install tilted\n```\n\n## Basic Usage\nTilted can be used to generate, evaluate and compare 5-card poker hands.\n\n### Hand Generation\nTo generate a random hand:\n```python\nfrom tilted import Deck, Hand\n\n\ndeck = Deck()\nunknown_hand = Hand(deck.draw_many(5))\n\nunknown_hand  # <Hand: 8♦ T♠ T♣ Q♣ K♥>\nunknown_hand.hand_rank  # <HandRank.PAIR: 2>\n```\n\n### Hand Evaluation\nTo evaluate an unknown hand:\n```python\nfrom tilted import Card, CardRank, CardSuit, Hand\n\n\nunknown_hand = Hand([\n  Card(CardRank.TEN, CardSuit.SPADES),\n  Card(CardRank.JACK, CardSuit.SPADES),\n  Card(CardRank.QUEEN, CardSuit.SPADES),\n  Card(CardRank.KING, CardSuit.SPADES),\n  Card(CardRank.ACE, CardSuit.SPADES),\n])\n\nunknown_hand.hand_rank  # <HandRank.ROYAL_FLUSH: 10>\n```\n\n### Hand Comparison\nTo compare two hands:\n```python\nfrom tilted import Card, CardRank, CardSuit, Hand\n\n\nroyal_flush = Hand([\n  Card(CardRank.TEN, CardSuit.SPADES),\n  Card(CardRank.JACK, CardSuit.SPADES),\n  Card(CardRank.QUEEN, CardSuit.SPADES),\n  Card(CardRank.KING, CardSuit.SPADES),\n  Card(CardRank.ACE, CardSuit.SPADES),\n])\n\nstraight_flush = Hand([\n  Card(CardRank.NINE, CardSuit.HEARTS),\n  Card(CardRank.TEN, CardSuit.HEARTS),\n  Card(CardRank.JACK, CardSuit.HEARTS),\n  Card(CardRank.QUEEN, CardSuit.HEARTS),\n  Card(CardRank.KING, CardSuit.HEARTS),\n])\n\nroyal_flush > straight_flush  # True\n```\n\n\n## Roadmap\n\n### Features\n- [x] Hand evaluation\n- [x] Hand comparison\n- [x] Deck support\n- [ ] Board & dealing (Flop, Turn, River)\n- [ ] Full board & hole card evaluation\n- [ ] Expand beyond Texas Hold \'Em (Pot Limit Omaha)\n\n### Deployment\n- [ ] CI\n',
    'author': 'Max Atkinson',
    'author_email': 'hiremaxatkinson@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MaxAtkinson/tilted',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, Final, Tuple

from tilted.enums import CardRank, HandRank

# Hand values based on the binary arithmetic
# FLUSH & STRAIGHT_FLUSH have no entries here
# as the methods for detection of flushy hands
# don't require binary arithmetic.
HIGH_CARD_VALUE: Final[int] = 5
PAIR_VALUE: Final[int] = 6
TWO_PAIR_VALUE: Final[int] = 7
THREE_OF_A_KIND_VALUE: Final[int] = 9
STRAIGHT_VALUE: Final[int] = 31
WHEEL_STRAIGHT_VALUE: Final[int] = 16444
FULL_HOUSE_VALUE: Final[int] = 10
FOUR_OF_A_KIND_VALUE: Final[int] = 1
ROYAL_FLUSH_VALUE: Final[int] = 31744

# Map of binary arithmetic results to hand ranks
# for easy access.
BINARY_ARITHMETIC_RESULT_TO_HAND_RANK: Final[Dict[int, HandRank]] = {
    HIGH_CARD_VALUE: HandRank.HIGH_CARD,
    PAIR_VALUE: HandRank.PAIR,
    TWO_PAIR_VALUE: HandRank.TWO_PAIR,
    THREE_OF_A_KIND_VALUE: HandRank.THREE_OF_A_KIND,
    STRAIGHT_VALUE: HandRank.STRAIGHT,
    WHEEL_STRAIGHT_VALUE: HandRank.STRAIGHT,
    FULL_HOUSE_VALUE: HandRank.FULL_HOUSE,
    FOUR_OF_A_KIND_VALUE: HandRank.FOUR_OF_A_KIND,
    ROYAL_FLUSH_VALUE: HandRank.ROYAL_FLUSH,
}


# We use 15 bits for binary arithmetic, but there
# are only 13 card ranks.
NUM_BITS: Final[int] = 15
UNUSED_BITS: Final[int] = NUM_BITS - len(CardRank)

# For ties, we sort the cards then left bit shift
# each card by the below numbers
TIEBREAKER_BITWISE_LEFT_SHIFTS: Final[Tuple[int, ...]] = (16, 12, 8, 4, 0)
